#include <QApplication>
#include <QCryptographicHash>
#include <QDebug>
#include <QEvent>
#include <QFileDialog>
#include <QLabel>
#include <QProxyStyle>
#include <QSettings>
#include <QStyleFactory>
#include <QTranslator>
#include "MainWindow.hxx"
#include "ui_MainWindow.h"
#include "filesystem/NCFilesystem.hxx"
#include "ImageWindow.hxx"
#include "TextWindow.hxx"
#include "DatWindow.hxx"
#include "ObjectWindow.hxx"

// --- internal typedefs ---

using QTW = QTableWidget;
using QTWI = QTableWidgetItem;
using QCH = QCryptographicHash;
using ColType = NC::FileData::Type;
using FileType = NC::Identify::Type;

// --- internal constants ---

const QColor Green(180, 255, 180);
const QColor Grey(180, 180, 180);
const QColor Red(255, 180, 180);
const QString TableHeader("header size;position;compressed size;original size;name length;name;SHA-1");
const QRegExp TextFiles("\\b(*.def|*.fx|*.ini|*.txt|.tsc|*.lua)\\b");

// --- constructors and deconstructors ---

MainWindow::MainWindow(QWidget *parent)
: QMainWindow(parent), _ui(new Ui::MainWindow), _translator(new QTranslator(this)),
  _settings(new QSettings("LinuxAddicted", "QTinNS Editor", this)), _vfs(new NCFilesystem), _img(new ImageWindow(this)),
  _txt(new TextWindow(this)), _dat(new DatWindow(this)), _obj(new ObjectWindow(this)), _def_status(new QLabel)
{
    _ui->setupUi(this);
    _ui->wid_table->setColumnCount(7);
    _ui->wid_table->setHorizontalHeaderLabels(TableHeader.split(";"));
    _ui->wid_table->setSelectionBehavior(QAbstractItemView::SelectRows);

    _ui->wid_stack->addWidget(_img);
    _ui->wid_stack->addWidget(_txt);
    _ui->wid_stack->addWidget(_dat);
    _ui->wid_stack->addWidget(_obj);
    _ui->wid_stack->setCurrentIndex(0);

    statusBar()->insertWidget(0, _def_status);

    setLanguage(Language::English);
    setupActions();

    if (_settings)
    {
        move(_settings->value("mw_position", QPoint(40, 40)).toPoint());
        resize(_settings->value("mw_size", QSize(800, 600)).toSize());
        setMinimumSize(_settings->value("mw_minsize", QSize(640, 480)).toSize());
    }
    else
    {
        move(40, 40);
        resize(800, 600);
        setMinimumSize(640, 480);
    }

    slotStatusUpdate();
    setMode(Mode::Fresh);
}

MainWindow::~MainWindow()
{
    if (_settings)
    {
        _settings->setValue("mw_position", pos());
        _settings->setValue("mw_size", size());
        _settings->setValue("mw_minsize", minimumSize());
        _settings->sync();
    }

    statusBar()->removeWidget(_def_status);

    _obj->setVisible(false);
    _ui->wid_stack->removeWidget(_obj);

    _dat->setVisible(false);
    _ui->wid_stack->removeWidget(_dat);

    _txt->setVisible(false);
    _ui->wid_stack->removeWidget(_txt);

    _img->setVisible(false);
    _ui->wid_stack->removeWidget(_img);

    delete _def_status;
    delete _obj;
    delete _dat;
    delete _txt;
    delete _img;
    delete _vfs;
    delete _settings;
    delete _translator;
    delete _ui;
}

// --- slots ---

void MainWindow::slotAppNew()
{
    _vfs->clear();
    setMode(Mode::Fresh);
}

void MainWindow::slotAppOpen()
{
    QString filename = QFileDialog::getOpenFileName(this, tr("DIA-OPEN-VFS"), "./", tr("FLT-FILE-VFS"));

    if (_vfs->load(filename) != _vfs->Error)
    {
        _filename = filename;
        slotStatusUpdate();
        setMode(Mode::Edited);
    }
}

void MainWindow::slotAppSave()
{
    if (_filename.isEmpty())
        slotAppSaveAs();
    else
        _vfs->save(_filename);
}

void MainWindow::slotAppSaveAs()
{
    QString filename = QFileDialog::getSaveFileName(this, tr("DIA-SAVE-VFS"), "./unnamed.pak", tr("FLT-FILE-VFS"));

    if (filename.size())
    {
        _filename = filename;
        _vfs->save(_filename);
    }
}

void MainWindow::slotAppQuit()
{
    close();
}

void MainWindow::slotFileAdd()
{
    QString filename = QFileDialog::getOpenFileName(this, tr("DIA-OPEN-FILE"), "./", tr("FLT-FILE-ANY"));

    if (filename.size())
    {
        _vfs->fileAdd(filename);
        slotStatusUpdate();
    }
}

void MainWindow::slotFileReplace()
{
    QString filename = QFileDialog::getOpenFileName(this, tr("DIA-OPEN-FILE"), "./", tr("FLT-FILE-ANY"));

    if (filename.size())
    {
        _vfs->fileReplace(filename, _ui->wid_table->currentRow());
        slotStatusUpdate();
    }
}

void MainWindow::slotFileRemove()
{
    _vfs->fileRemove(_ui->wid_table->currentRow());
    slotStatusUpdate();
    setMode(Mode::Edited);
}

void MainWindow::slotFileSavePlain()
{
    QString filename = QFileDialog::getSaveFileName(this, tr("DIA-SAVE-PLAIN"), "./file", tr("FLT-FILE-PLAIN"));

    if (filename.size())
        _vfs->fileSave(filename, _ui->wid_table->currentRow());
}

void MainWindow::slotFileSavePAK()
{
    QString filename = QFileDialog::getSaveFileName(this, tr("DIA-SAVE-PAK"), "./file", tr("FLT-FILE-PAK"));

    if (filename.size())
        _vfs->fileSave(filename, _ui->wid_table->currentRow(), FileType::PAK1);
}

void MainWindow::slotFileSaveZLIB()
{
    QString filename = QFileDialog::getSaveFileName(this, tr("DIA-SAVE-ZLIB"), "./file", tr("FLT-FILE-ZLIB"));

    if (filename.size())
        _vfs->fileSave(filename, _ui->wid_table->currentRow(), FileType::ZLIB);
}

void MainWindow::slotFileConvWin()
{
    QTW *table = _ui->wid_table;
    NCFile entry = _vfs->fileGet(table->currentRow());

    entry.setName(entry.name().replace("/", "\\"));
    _vfs->fileReplace(entry, table->currentRow());
    slotTableSelected();
}

void MainWindow::slotFileConvLin()
{
    QTW *table = _ui->wid_table;
    NCFile entry = _vfs->fileGet(table->currentRow());

    entry.setName(entry.name().replace("\\", "/"));
    _vfs->fileReplace(entry, table->currentRow());
    slotTableSelected();
}

void MainWindow::slotFileImage()
{
    QTW *table = _ui->wid_table;
    NCFile entry = _vfs->fileGet(table->currentRow());

    _img->slotImageView(entry, table->currentRow());
    _ui->wid_stack->setCurrentWidget(_img);
    setMode(Mode::Blocked);
}

void MainWindow::slotFileText()
{
    QTW *table = _ui->wid_table;
    NCFile entry = _vfs->fileGet(table->currentRow());

    _txt->slotTextView(entry, table->currentRow());
    _ui->wid_stack->setCurrentWidget(_txt);
    setMode(Mode::Blocked);
}

void MainWindow::slotFileDat()
{
    QTW *table = _ui->wid_table;
    NCFile entry = _vfs->fileGet(table->currentRow());

    _dat->slotDatView(entry, table->currentRow());
    _ui->wid_stack->setCurrentWidget(_dat);
    setMode(Mode::Blocked);
}

void MainWindow::slotFileObject()
{
    QTW *table = _ui->wid_table;
    NCFile entry = _vfs->fileGet(table->currentRow());

    _obj->slotObjectView(entry, table->currentRow());
    _ui->wid_stack->setCurrentWidget(_obj);
    setMode(Mode::Blocked);
}

void MainWindow::slotLangEnglish()
{
    setLanguage(Language::English);
}

void MainWindow::slotLangFrench()
{
    setLanguage(Language::French);
}

void MainWindow::slotLangGerman()
{
    setLanguage(Language::German);
}

void MainWindow::slotStyleDefault()
{
    QApplication::setStyle(new QProxyStyle);
}

void MainWindow::slotStyleFusion()
{
    QApplication::setStyle(QStyleFactory::create("Fusion"));
}

void MainWindow::slotStyleWindows()
{
    QApplication::setStyle(QStyleFactory::create("Windows"));
}

void MainWindow::slotStyleGTK()
{
    QApplication::setStyle(QStyleFactory::create("GTK+"));
}

void MainWindow::slotAboutApp()
{
}

void MainWindow::slotAboutQt()
{
    qApp->aboutQt();
}

void MainWindow::slotStatusUpdate()
{
    QString msg1 = " " + tr("MSG-UPD-FILE") + ": '" + _filename + "' " + tr("MSG-UPD-WITH") + " "
                   + QString::number(_vfs->count()) + " " + tr("MSG-UPD-ENTRIES") + " ";

    _def_status->setText(msg1);
}

void MainWindow::slotTableSelected()
{
    QTW *table = _ui->wid_table;
    NCFile entry = _vfs->fileGet(table->currentRow());
    QImage image = QImage::fromData(entry.oData());
    QString text = entry.name();

    setMode(Mode::Selected);

    if (entry.name().count("/"))
        _ui->act_file_conv_win->setEnabled(true);
    else
        _ui->act_file_conv_win->setEnabled(false);

    if (entry.name().count("\\"))
        _ui->act_file_conv_lin->setEnabled(true);
    else
        _ui->act_file_conv_lin->setEnabled(false);

    if (image.isNull())
        _ui->act_file_image->setEnabled(false);
    else
        _ui->act_file_image->setEnabled(true);

    if (text.contains(".def", Qt::CaseInsensitive) || text.contains(".fx", Qt::CaseInsensitive)
        || text.contains(".ini", Qt::CaseInsensitive) || text.contains(".txt", Qt::CaseInsensitive)
        || text.contains(".tsc", Qt::CaseInsensitive) || text.contains(".lua", Qt::CaseInsensitive))
        _ui->act_file_text->setEnabled(true);
    else
        _ui->act_file_image->setEnabled(false);

    if (text.contains(".dat", Qt::CaseInsensitive))
        _ui->act_file_dat->setEnabled(true);
    else
        _ui->act_file_dat->setEnabled(false);

    if (text.contains(".bsp", Qt::CaseInsensitive))
        _ui->act_file_object->setEnabled(true);
    else
        _ui->act_file_object->setEnabled(false);
}

void MainWindow::slotTableChange(qint32 row)
{
    QTW *table = _ui->wid_table;
    NCFile entry = _vfs->fileGet(row);

    entry.setName(table->item(row, 5)->text());
    entry.setNameLength(entry.name().size() + 1);
    entry.setHSize(entry.headerSize() + entry.nameLength());

    _vfs->fileReplace(entry, row);
    slotStatusUpdate();
}

void MainWindow::slotTableStatus(quint32 from_row, quint32 to_row, ColType cols)
{
    QTW *table = _ui->wid_table;
    QTWI *it0 = 0;
    QTWI *it1 = 0;
    QTWI *it2 = 0;
    QTWI *it3 = 0;
    QTWI *it4 = 0;
    QTWI *it5 = 0;
    QTWI *it6 = 0;

    // if that happens, something is very very odd
    if (to_row != _vfs->count())
        qDebug() << "it happend! oh my god, it happend!!";

    // if these do not fit, we have a full refresh and have to build every single
    // table item
    if (to_row != static_cast<quint32>(table->rowCount()))
    {
        disconnect(table, SIGNAL(cellChanged(qint32, qint32)),
                   this, SLOT(slotTableChange(qint32)));

        table->setRowCount(to_row);
        for (quint32 row = 0; row < to_row; ++row)
        {
            NCFile entry = _vfs->fileGet(row);

            it0 = new QTWI(QString::number(entry.hSize()));
            it1 = new QTWI(QString::number(entry.position()));
            it2 = new QTWI(QString::number(entry.cSize()));
            it3 = new QTWI(QString::number(entry.oSize()));
            it4 = new QTWI(QString::number(entry.nameLength()));
            it5 = new QTWI(entry.name());
            it6 = new QTWI(QString(QCH::hash(entry.cData(), QCH::Sha1).toHex().toUpper()));

            it0->setBackground(Grey);
            it0->setFlags(it0->flags() ^ Qt::ItemIsEditable);
            it0->setTextAlignment(Qt::AlignRight);

            it1->setBackground(Grey);
            it1->setFlags(it1->flags() ^ Qt::ItemIsEditable);
            it1->setTextAlignment(Qt::AlignRight);

            it2->setBackground(Grey);
            it2->setFlags(it2->flags() ^ Qt::ItemIsEditable);
            it2->setTextAlignment(Qt::AlignRight);

            it3->setBackground(Grey);
            it3->setFlags(it3->flags() ^ Qt::ItemIsEditable);
            it3->setTextAlignment(Qt::AlignRight);

            it4->setBackground(Grey);
            it4->setFlags(it4->flags() ^ Qt::ItemIsEditable);
            it4->setTextAlignment(Qt::AlignRight);

            it5->setBackground(Green);
            it5->setTextAlignment(Qt::AlignLeft);

            it6->setBackground(Grey);
            it6->setFlags(it6->flags() ^ Qt::ItemIsEditable);
            it6->setTextAlignment(Qt::AlignLeft);

            switch (cols)
            {
                case ColType::HSize:
                    it0->setBackground(Red);
                    break;

                case ColType::Position:
                    it1->setBackground(Red);
                    break;

                case ColType::CSize:
                    it2->setBackground(Red);
                    break;

                case ColType::OSize:
                    it3->setBackground(Red);
                    break;

                case ColType::NameLength:
                    it4->setBackground(Red);
                    break;

                case ColType::Name:
                    it5->setBackground(Red);
                    break;

                case ColType::CData:
                case ColType::OData:
                    break;

                case ColType::All:
                    it0->setBackground(Red);
                    it1->setBackground(Red);
                    it2->setBackground(Red);
                    it3->setBackground(Red);
                    it4->setBackground(Red);
                    it5->setBackground(Red);
                    it6->setBackground(Red);
                    break;

                case ColType::None:
                default:
                    it0->setBackground(Grey);
                    it1->setBackground(Grey);
                    it2->setBackground(Grey);
                    it3->setBackground(Grey);
                    it4->setBackground(Grey);
                    it5->setBackground(Green);
                    it6->setBackground(Grey);
            }

            table->setItem(row, 0, it0);
            table->setItem(row, 1, it1);
            table->setItem(row, 2, it2);
            table->setItem(row, 3, it3);
            table->setItem(row, 4, it4);
            table->setItem(row, 5, it5);
            table->setItem(row, 6, it6);
        }
        connect(table, SIGNAL(cellChanged(qint32, qint32)),
                this, SLOT(slotTableChange(qint32)));
    }
    else
    {
        disconnect(table, SIGNAL(cellChanged(qint32, qint32)),
                   this, SLOT(slotTableChange(qint32)));

        for (quint32 row = 0; row < to_row; ++row)
        {
            NCFile entry = _vfs->fileGet(row);

            table->item(row, 1)->setText(QString::number(entry.position()));
        }

        for (quint32 row = from_row; row < to_row; ++row)
        {
            switch (cols)
            {
                case ColType::HSize:
                    table->item(row, 0)->setBackground(Red);
                    break;

                case ColType::Position:
                    table->item(row, 1)->setBackground(Red);
                    break;

                case ColType::CSize:
                    table->item(row, 2)->setBackground(Red);
                    break;

                case ColType::OSize:
                    table->item(row, 3)->setBackground(Red);
                    break;

                case ColType::NameLength:
                    table->item(row, 4)->setBackground(Red);
                    break;

                case ColType::Name:
                    table->item(row, 5)->setBackground(Red);
                    break;

                case ColType::CData:
                case ColType::OData:
                    break;

                case ColType::All:
                    table->item(row, 0)->setBackground(Red);
                    table->item(row, 1)->setBackground(Red);
                    table->item(row, 2)->setBackground(Red);
                    table->item(row, 3)->setBackground(Red);
                    table->item(row, 4)->setBackground(Red);
                    table->item(row, 5)->setBackground(Red);
                    table->item(row, 6)->setBackground(Red);
                    break;

                case ColType::None:
                default:
                    table->item(row, 0)->setBackground(Grey);
                    table->item(row, 1)->setBackground(Grey);
                    table->item(row, 2)->setBackground(Grey);
                    table->item(row, 3)->setBackground(Grey);
                    table->item(row, 4)->setBackground(Grey);
                    table->item(row, 5)->setBackground(Green);
                    table->item(row, 6)->setBackground(Grey);
            }
        }

        connect(table, SIGNAL(cellChanged(qint32, qint32)),
                this, SLOT(slotTableChange(qint32)));
    }

    table->resizeColumnsToContents();
}

void MainWindow::slotTableAdd(NCFile& entry, quint32 row)
{
    QTW *table = _ui->wid_table;
    QTWI *it0 = new QTWI(QString::number(entry.hSize()));
    QTWI *it1 = new QTWI(QString::number(entry.position()));
    QTWI *it2 = new QTWI(QString::number(entry.cSize()));
    QTWI *it3 = new QTWI(QString::number(entry.oSize()));
    QTWI *it4 = new QTWI(QString::number(entry.nameLength()));
    QTWI *it5 = new QTWI(entry.name());
    QTWI *it6 = new QTWI(QString(QCH::hash(entry.cData(), QCH::Sha1).toHex().toUpper()));

    it0->setBackground(Red);
    it0->setFlags(it0->flags() ^ Qt::ItemIsEditable);
    it0->setTextAlignment(Qt::AlignRight);

    it1->setBackground(Red);
    it1->setFlags(it1->flags() ^ Qt::ItemIsEditable);
    it1->setTextAlignment(Qt::AlignRight);

    it2->setBackground(Red);
    it2->setFlags(it2->flags() ^ Qt::ItemIsEditable);
    it2->setTextAlignment(Qt::AlignRight);

    it3->setBackground(Red);
    it3->setFlags(it3->flags() ^ Qt::ItemIsEditable);
    it3->setTextAlignment(Qt::AlignRight);

    it4->setBackground(Red);
    it4->setFlags(it4->flags() ^ Qt::ItemIsEditable);
    it4->setTextAlignment(Qt::AlignRight);

    it5->setBackground(Red);
    it5->setTextAlignment(Qt::AlignLeft);

    it6->setBackground(Red);
    it6->setFlags(it6->flags() ^ Qt::ItemIsEditable);
    it6->setTextAlignment(Qt::AlignLeft);

    disconnect(table, SIGNAL(cellChanged(qint32, qint32)),
               this, SLOT(slotTableChange(qint32)));

    table->insertRow(row);
    table->setItem(row, 0, it0);
    table->setItem(row, 1, it1);
    table->setItem(row, 2, it2);
    table->setItem(row, 3, it3);
    table->setItem(row, 4, it4);
    table->setItem(row, 5, it5);
    table->setItem(row, 6, it6);

    connect(table, SIGNAL(cellChanged(qint32, qint32)),
            this, SLOT(slotTableChange(qint32)));

    setMode(Mode::Edited);
}

void MainWindow::slotTableReplace(NCFile& entry, quint32 row)
{
    QTW *table = _ui->wid_table;

    disconnect(table, SIGNAL(cellChanged(qint32, qint32)),
               this, SLOT(slotTableChange(qint32)));

    table->item(row, 0)->setText(QString::number(entry.hSize()));
    table->item(row, 0)->setBackground(Red);

    table->item(row, 1)->setText(QString::number(entry.position()));
    table->item(row, 1)->setBackground(Red);

    table->item(row, 2)->setText(QString::number(entry.cSize()));
    table->item(row, 2)->setBackground(Red);

    table->item(row, 3)->setText(QString::number(entry.oSize()));
    table->item(row, 3)->setBackground(Red);

    table->item(row, 4)->setText(QString::number(entry.nameLength()));
    table->item(row, 4)->setBackground(Red);

    table->item(row, 5)->setText(entry.name());
    table->item(row, 5)->setBackground(Red);

    table->item(row, 6)->setText(QString(QCH::hash(entry.cData(), QCH::Sha1).toHex().toUpper()));
    table->item(row, 6)->setBackground(Red);

    connect(table, SIGNAL(cellChanged(qint32, qint32)),
            this, SLOT(slotTableChange(qint32)));

    _ui->wid_stack->setCurrentIndex(0);
    setMode(Mode::Edited);
}

void MainWindow::slotTableRemove(quint32 row)
{
    _ui->wid_table->removeRow(row);
    setMode(Mode::Edited);
}

void MainWindow::slotEntryReplace(NCFile& entry, quint32 row)
{
    _vfs->fileReplace(entry, row);
}

void MainWindow::slotEntryIgnore()
{
    _ui->wid_stack->setCurrentIndex(0);
    setMode(Mode::Edited);
}

// --- protected methods ---

void MainWindow::changeEvent(QEvent *event)
{
    if (event->type() == QEvent::LanguageChange)
        _ui->retranslateUi(this);
    else
        QMainWindow::changeEvent(event);
}

void MainWindow::setLanguage(const Language lang)
{
    QTranslator *translator = new QTranslator(this);

    if (translator)
    {
        switch (lang)
        {
            case Language::French:
                translator->load(":/Language/French");
                break;
            case Language::German:
                translator->load(":/Language/German");
                break;
            case Language::English:
            default:
                translator->load(":/Language/English");
        }

        if (!translator->isEmpty())
        {
            qApp->removeTranslator(_translator);
            delete _translator;
            _translator = translator;
            qApp->installTranslator(_translator);

            _ui->act_lang_english->setEnabled(true);
            _ui->act_lang_french->setEnabled(true);
            _ui->act_lang_german->setEnabled(true);

            switch (lang)
            {
                case Language::French:
                    _ui->act_lang_french->setEnabled(false);
                    setLocale({"fr_FR.UTF-8"});
                    break;
                case Language::German:
                    _ui->act_lang_german->setEnabled(false);
                    setLocale({"de_DE.UTF-8"});
                    break;
                case Language::English:
                default:
                    _ui->act_lang_english->setEnabled(false);
                    setLocale({"us_EN.UTF-8"});
            }
        }
        else
            delete translator;
    }
}

void MainWindow::setMode(const Mode mode)
{
    _ui->act_app_new->setEnabled(false);
    _ui->act_app_open->setEnabled(false);
    _ui->act_app_save->setEnabled(false);
    _ui->act_app_saveas->setEnabled(false);
    _ui->act_app_quit->setEnabled(false);

    _ui->mm_file->setEnabled(false);

    _ui->act_file_add->setEnabled(false);
    _ui->act_file_replace->setEnabled(false);
    _ui->act_file_remove->setEnabled(false);
    _ui->act_file_save_plain->setEnabled(false);
    _ui->act_file_save_pak->setEnabled(false);
    _ui->act_file_save_zlib->setEnabled(false);
    _ui->act_file_conv_win->setEnabled(false);
    _ui->act_file_conv_lin->setEnabled(false);
    _ui->act_file_image->setEnabled(false);
    _ui->act_file_text->setEnabled(false);
    _ui->act_file_dat->setEnabled(false);
    _ui->act_file_object->setEnabled(false);

    switch (mode)
    {
        case Mode::Fresh:
            _filename = "";

            _ui->act_app_open->setEnabled(true);
            _ui->act_app_quit->setEnabled(true);

            _ui->mm_file->setEnabled(true);

            _ui->act_file_add->setEnabled(true);

            break;

        case Mode::Edited:
            _ui->wid_table->clearSelection();

            _ui->act_app_new->setEnabled(true);
            _ui->act_app_save->setEnabled(true);
            _ui->act_app_saveas->setEnabled(true);
            _ui->act_app_quit->setEnabled(true);

            _ui->mm_file->setEnabled(true);

            _ui->act_file_add->setEnabled(true);

            break;

        case Mode::Selected:
            _ui->act_app_new->setEnabled(true);
            _ui->act_app_save->setEnabled(true);
            _ui->act_app_saveas->setEnabled(true);
            _ui->act_app_quit->setEnabled(true);

            _ui->mm_file->setEnabled(true);

            _ui->act_file_add->setEnabled(true);
            _ui->act_file_replace->setEnabled(true);
            _ui->act_file_remove->setEnabled(true);
            _ui->act_file_save_plain->setEnabled(true);
            _ui->act_file_save_pak->setEnabled(true);
            _ui->act_file_save_zlib->setEnabled(true);

            break;

        case Mode::Blocked:
            _ui->act_app_quit->setEnabled(true);
    }
}

void MainWindow::setupActions()
{
    QTW *table = _ui->wid_table;
    QAction *ctx_sep1 = new QAction(this);
    QAction *ctx_sep2 = new QAction(this);
    QAction *ctx_sep3 = new QAction(this);
    QAction *ctx_sep4 = new QAction(this);

    // all the nice connections we need for the GUI
    connect(_ui->act_app_new, SIGNAL(triggered()),
            this, SLOT(slotAppNew()));
    connect(_ui->act_app_open, SIGNAL(triggered()),
            this, SLOT(slotAppOpen()));
    connect(_ui->act_app_save, SIGNAL(triggered()),
            this, SLOT(slotAppSave()));
    connect(_ui->act_app_saveas, SIGNAL(triggered()),
            this, SLOT(slotAppSaveAs()));
    connect(_ui->act_app_quit, SIGNAL(triggered()),
            this, SLOT(slotAppQuit()));

    connect(_ui->act_file_add, SIGNAL(triggered()),
            this, SLOT(slotFileAdd()));
    connect(_ui->act_file_replace, SIGNAL(triggered()),
            this, SLOT(slotFileReplace()));
    connect(_ui->act_file_remove, SIGNAL(triggered()),
            this, SLOT(slotFileRemove()));
    connect(_ui->act_file_save_plain, SIGNAL(triggered()),
            this, SLOT(slotFileSavePlain()));
    connect(_ui->act_file_save_pak, SIGNAL(triggered()),
            this, SLOT(slotFileSavePAK()));
    connect(_ui->act_file_save_zlib, SIGNAL(triggered()),
            this, SLOT(slotFileSaveZLIB()));
    connect(_ui->act_file_conv_win, SIGNAL(triggered()),
            this, SLOT(slotFileConvWin()));
    connect(_ui->act_file_conv_lin, SIGNAL(triggered()),
            this, SLOT(slotFileConvLin()));
    connect(_ui->act_file_image, SIGNAL(triggered()),
            this, SLOT(slotFileImage()));
    connect(_ui->act_file_text, SIGNAL(triggered()),
            this, SLOT(slotFileText()));
    connect(_ui->act_file_dat, SIGNAL(triggered()),
            this, SLOT(slotFileDat()));
    connect(_ui->act_file_object, SIGNAL(triggered()),
        this, SLOT(slotFileObject()));

    connect(_ui->act_lang_english, SIGNAL(triggered()),
            this, SLOT(slotLangEnglish()));
    connect(_ui->act_lang_french, SIGNAL(triggered()),
            this, SLOT(slotLangFrench()));
    connect(_ui->act_lang_german, SIGNAL(triggered()),
            this, SLOT(slotLangGerman()));

    connect(_ui->act_style_default, SIGNAL(triggered()),
            this, SLOT(slotStyleDefault()));
    connect(_ui->act_style_fusion, SIGNAL(triggered()),
            this, SLOT(slotStyleFusion()));
    connect(_ui->act_style_windows, SIGNAL(triggered()),
            this, SLOT(slotStyleWindows()));
    connect(_ui->act_style_gtk, SIGNAL(triggered()),
            this, SLOT(slotStyleGTK()));

    connect(_ui->act_about_app, SIGNAL(triggered()),
            this, SLOT(slotAboutApp()));
    connect(_ui->act_about_qt, SIGNAL(triggered()),
            this, SLOT(slotAboutQt()));

    // table update connections
    connect(table, SIGNAL(cellPressed(qint32, qint32)),
            this, SLOT(slotTableSelected()));
    connect(table, SIGNAL(cellChanged(qint32, qint32)),
            this, SLOT(slotTableChange(qint32)));

    connect(_vfs, SIGNAL(sigFileStatus(quint32, quint32, ColType)),
            this, SLOT(slotTableStatus(quint32, quint32, ColType)));
    connect(_vfs, SIGNAL(sigFileAdd(NCFile&, quint32)),
            this, SLOT(slotTableAdd(NCFile&, quint32)));
    connect(_vfs, SIGNAL(sigFileReplace(NCFile&, quint32)),
            this, SLOT(slotTableReplace(NCFile&, quint32)));
    connect(_vfs, SIGNAL(sigFileRemove(quint32)),
            this, SLOT(slotTableRemove(quint32)));

    // image viewer connections
    connect(_img, SIGNAL(sigImageUpdate(NCFile&, quint32)),
            this, SLOT(slotEntryReplace(NCFile&, quint32)));
    connect(_img, SIGNAL(sigImageIgnored()),
            this, SLOT(slotEntryIgnore()));

    // text viewer connections
    connect(_txt, SIGNAL(sigTextUpdate(NCFile&, quint32)),
            this, SLOT(slotEntryReplace(NCFile&, quint32)));
    connect(_txt, SIGNAL(sigTextIgnored()),
            this, SLOT(slotEntryIgnore()));

    // dat viewer connections
    connect(_dat, SIGNAL(sigDatUpdate(NCFile&, quint32)),
            this, SLOT(slotEntryReplace(NCFile&, quint32)));
    connect(_dat, SIGNAL(sigDatIgnored()),
            this, SLOT(slotEntryIgnore()));

    // object viewer connections
    connect(_obj, SIGNAL(sigObjectUpdate(NCFile&, quint32)),
            this, SLOT(slotEntryReplace(NCFile&, quint32)));
    connect(_obj, SIGNAL(sigObjectIgnored()),
            this, SLOT(slotEntryIgnore()));

    // context menus for the QTableWidget
    ctx_sep1->setSeparator(true);
    ctx_sep2->setSeparator(true);
    ctx_sep3->setSeparator(true);
    ctx_sep4->setSeparator(true);

    table->addAction(_ui->act_file_add);
    table->addAction(ctx_sep1);
    table->addAction(_ui->act_file_replace);
    table->addAction(_ui->act_file_remove);
    table->addAction(ctx_sep2);
    table->addAction(_ui->act_file_save_plain);
    table->addAction(_ui->act_file_save_pak);
    table->addAction(_ui->act_file_save_zlib);
    table->addAction(ctx_sep3);
    table->addAction(_ui->act_file_conv_win);
    table->addAction(_ui->act_file_conv_lin);
    table->addAction(ctx_sep4);
    table->addAction(_ui->act_file_image);
    table->addAction(_ui->act_file_text);
    table->addAction(_ui->act_file_dat);
    table->addAction(_ui->act_file_object);
}
