#pragma once

#include "common/BaseObject.hxx"
#include "Vector.hxx"
#include "Ray.hxx"

enum class ObjectType : uint8_t {
    None,
    Camera,
    Plane,
    Sphere,
    Mesh
};

template <typename MT>
class Object : public BaseObject<Object<MT>> {
    static_assert (std::is_integral<MT>::value || std::is_floating_point<MT>::value,
                   "ERROR: template parameter is not an integral or floating point type");
public:
    // --- constructors and deconstructors ---

    Object()
    : BaseObject<Object<MT>>(), _name("unknown"), _visible(true), _origin(0.0, 0.0, 0.0), _up(0.0, 1.0, 0.0)
    {
    }

    Object(const std::string& name, const bool vis, const Vector3<MT>& org, const Vector3<MT>& up)
    : BaseObject<Object<MT>>(), _name(name), _visible(vis), _origin(org), _up(up)
    {
    }

    Object(const Object& rhs)
    : BaseObject<Object<MT>>(rhs), _name(rhs.name), _visible(rhs.visible), _origin(rhs._origin), _up(rhs._up)
    {
    }

    Object(Object&& rhs)
    : BaseObject<Object<MT>>(rhs), _name(std::move(rhs._name)), _visible(std::move(rhs._visible)),
      _origin(std::move(rhs._origin)), _up(std::move(rhs._up))
    {
    }

    virtual ~Object()
    {
    }

    // --- default operators ---

    Object& operator=(const Object& rhs)
    {
        if (this != &rhs)
        {
            BaseObject<Object<MT>>::operator=(rhs);

            _name = rhs._name;
            _visible = rhs._visible;
            _origin = rhs._origin;
            _up = rhs._up;
        }

        return *this;
    }

    Object& operator=(Object&& rhs)
    {
        if (this != &rhs)
        {
            BaseObject<Object<MT>>::operator=(rhs);

            _name = std::move(rhs._name);
            _visible = std::move(rhs._visible);
            _origin = std::move(rhs._origin);
            _up = std::move(rhs._up);
        }

        return *this;
    }

    // --- public methods ---

    virtual MT intersect(const Ray<MT>& /*ray*/) const
    {
        return -1.0;
    }

    virtual std::string name() const
    {
        return _name;
    }

    virtual Vector3<MT> origin() const
    {
        return _origin;
    }

    virtual void setName(const std::string& name)
    {
        _name = name;
    }

    virtual void setOrigin(const Vector3<MT>& o)
    {
        _origin = o;
    }

    virtual void setUp(const Vector3<MT>& up)
    {
        _up = up;
    }

    virtual void setVisible(const bool visible)
    {
        _visible = visible;
    }

    virtual ObjectType type() const
    {
        return ObjectType::None;
    }

    virtual Vector3<MT> up() const
    {
        return _up;
    }

    virtual bool visible() const
    {
        return _visible;
    }

protected:
    // --- protected data structures ---

    std::string _name;
    bool _visible;
    Vector3<MT> _origin;
    Vector3<MT> _up;
};
