#pragma once

#include "Vector.hxx"

template <typename MT>
class Ray {
    static_assert (std::is_integral<MT>::value || std::is_floating_point<MT>::value,
                   "ERROR: template parameter is not an integral or floating point type");
public:
    // --- constructors and deconstructors ---

    Ray()
    : _origin(0.0, 0.0, 0.0), _direction(0.0, 0.0, 0.0), _min(1.0), _max(1000.0), _depth(10)
    {
    }

    Ray(const Vector3<MT>& org, const Vector3<MT>& dir, const MT& min, const MT& max, const size_t& depth)
    : _origin(org), _direction(dir), _min(min), _max(max), _depth(depth)
    {
    }

    Ray(const Ray& rhs)
    : _origin(rhs._origin), _direction(rhs._direction), _min(rhs._min), _max(rhs._max), _depth(rhs._depth)
    {
    }

    Ray(Ray&& rhs)
    : _origin(std::move(rhs._origin)), _direction(std::move(rhs._direction)), _min(std::move(rhs._min)),
      _max(std::move(rhs._max)), _depth(std::move(rhs._depth))
    {
    }

    virtual ~Ray()
    {
    }

    // --- default operators ---

    Ray& operator=(const Ray& rhs)
    {
        if (this != &rhs)
        {
            _origin = rhs._origin;
            _direction = rhs._direction;
            _min = rhs._min;
            _max = rhs._max;
            _depth = rhs._depth;
        }

        return *this;
    }

    Ray& operator=(Ray& rhs)
    {
        if (this != &rhs)
        {
            _origin = std::move(rhs._origin);
            _direction = std::move(rhs._direction);
            _min = std::move(rhs._min);
            _max = std::move(rhs._max);
            _depth = std::move(rhs._depth);
        }

        return *this;
    }

    bool operator==(const Ray& rhs) const
    {
        return _origin == rhs._origin &&
               _direction == rhs._direction &&
               _min == rhs._min &&
               _max == rhs._max &&
               _depth == rhs._depth;
    }

    bool operator!=(const Ray& rhs) const
    {
        return !(*this == rhs);
    }

    // --- public methods ---

    size_t depth() const
    {
        return _depth;
    }

    Vector3<MT> direction() const
    {
        return _direction;
    }

    MT max() const
    {
        return _max;
    }

    MT min() const
    {
        return _min;
    }

    Vector3<MT> origin() const
    {
        return _origin;
    }

    void setDepth(const size_t& depth)
    {
        _depth = depth;
    }

    void setDirection(const Vector3<MT>& dir)
    {
        _direction = dir.normalized();
    }

    void setMax(const MT& max)
    {
        _max = max;
    }

    void setMin(const MT& min)
    {
        _min = min;
    }

    void setOrigin(const Vector3<MT>& org)
    {
        _origin = org;
    }

    std::string string() const
    {
        std::stringstream result;

        result << "org(" << _origin.string()
               << "), dir(" << _direction.string()
               << "), min(" << _min
               << "), max(" << _max
               << "), dep(" << _depth
               << ")";

        return result.str();
    }

protected:
    // --- protected data structures ---

    Vector3<MT> _origin;
    Vector3<MT> _direction;
    MT _min;                // near plane value
    MT _max;                // far plane value
    size_t _depth;          // recursion depth
};
